
import logging

from brownie import chain
from dank_mids import instances
from dank_mids.loggers import demo_logger
from multicall import Call
from multicall.utils import await_awaitable, gather

from tests.fixtures import dank_w3

demo_logger.addHandler(logging.StreamHandler())
demo_logger.setLevel(logging.DEBUG)

CHAI = '0x06AF07097C9Eeb7fD685c692751D5C66dB49c215'
BIG_WORK = [Call(CHAI, 'totalSupply()(uint)', [[f'totalSupply{i}',None]],_w3=dank_w3).coroutine() for i in range(100_000)]
height = chain.height
MULTIBLOCK_WORK = [Call(CHAI, 'totalSupply()(uint)', [[f'totalSupply{i}',None]],_w3=dank_w3, block_id=height-i).coroutine() for i in range(100_000)]

def test_dank_middleware():
    await_awaitable(gather(BIG_WORK))
    dank_mids_controller = instances[0]
    cid = dank_mids_controller.call_uid.latest
    mid = dank_mids_controller.worker.multicall_uid.latest
    rid = dank_mids_controller.worker.request_uid.latest
    assert cid, "The DankMiddlewareController did not process any calls."
    assert mid, "The DankMiddlewareController did not process any batches."
    assert rid, "The DankMiddlewareController did not process any requests."
    print(f"calls:             {cid}")
    print(f"multicalls:        {rid}")
    print(f"batches:           {mid}")
    print(f"calls per batch:   {cid/mid}")
    print(f"calls per request:   {cid/rid}")
    print(f"requests per batch:  {rid/mid}")
    # General "tests" that verify batching performance
    assert mid < cid / 10, f"Batched {cid} calls into {mid} batches. Performance underwhelming."
    assert rid < cid / 5,  f"Batched {cid} calls into {rid} multicalls. Performance underwhelming."
    assert mid < rid / 2,  f"{mid} batches required {rid} multicalls. Performance underwhelming."

def test_bad_hex_handling():
    chainlinkfeed = "0xfe67209f6FE3BA6cE36d0941700085C194e958DF"
    assert await_awaitable(Call(chainlinkfeed, 'latestAnswer()(uint)', block_id=14_000_000).coroutine()) == 15717100
    assert chainlinkfeed in instances[0].do_not_batch

def test_next_cid():
    assert instances[0].call_uid.next + 1 == instances[0].call_uid.next
    
def test_next_mid():
    assert instances[0].worker.request_uid.next + 1 == instances[0].worker.request_uid.next
    
def test_next_bid():
    assert instances[0].worker.multicall_uid.next + 1 == instances[0].worker.multicall_uid.next

def test_json_batching():

    # reset uids
    dank_mids_controller = instances[0]
    dank_mids_controller.call_uid._value = 0
    dank_mids_controller.worker.request_uid._value = 0

    await_awaitable(gather(MULTIBLOCK_WORK))
    cid = dank_mids_controller.call_uid.latest
    jid = dank_mids_controller.worker.jsonrpc_batch_uid.latest
    rid = dank_mids_controller.worker.request_uid.latest
    assert jid, "The DankMiddlewareController did not process any jsonrpc batches."
    print(f"jsonrpc batches:    {jid}")
    print(f'requests:           {rid}')
    print(f'requests per batch: {rid / jid}')
    print(f'calls per request   {cid / rid}')
    print(f'calls per batch     {cid / jid}')



import asyncio
from typing import TYPE_CHECKING, Any, Generator, Optional, Union

from aiohttp import RequestInfo
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from multicall.utils import run_in_subprocess
from web3 import Web3
from web3.types import RPCError, RPCResponse

from dank_mids.constants import BAD_HEXES
from dank_mids.loggers import demo_logger, main_logger
from dank_mids.types import BlockId

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController


class ResponseNotReady(Exception):
    pass

def _call_failed(data: Optional[Union[bytes, Exception]]) -> bool:
    """ Returns True if `data` indicates a failed response, False otherwise. """
    if data is None:
        return True
    # If we got a known "bad" response from the multicall, that is also a failure.
    # Most likely the target contract does not support multicall2.
    elif (isinstance(data, bytes) and HexBytes(data).hex() in BAD_HEXES):
        return True
    return False

def _reattempt_call_and_return_exception(target: ChecksumAddress, calldata: bytes, block: BlockId, w3: Web3) -> Union[bytes, Exception]:
    """ NOTE: This runs synchronously in a subprocess in order to bypass Dank Middleware without blocking the event loop. """
    try:
        return w3.eth.call({"to": target, "data": calldata}, block)
    except Exception as e:
        return e

def _err_response(e: Exception) -> RPCError:
    """ Extract an error message from `e` to use in a spoof rpc response. """
    if isinstance(e.args[0], str) or isinstance(e.args[0], RequestInfo):
        err_msg = f"DankMidsError: {type(e)} {e.args}"
    elif isinstance(e.args[0], Exception):
        err_msg = f"DankMidsError: {type(e.args[0])} {e.args[0].args}"
    elif "message" in e.args[0]:
        err_msg = e.args[0]["message"]
    elif "error" in e.args[0] and "message" in e.args[0]["error"]:
        err_msg = e.args[0]["error"]["message"]
    else:
        raise e
    return {'code': -32000, 'message': err_msg, 'data': ''}


class BatchedCall:
    def __init__(self, controller: "DankMiddlewareController", params: Any) -> None:
        """ Adds a call to the DankMiddlewareContoller's `pending_calls`. """
        self.cid = controller.call_uid.next
        self.block: BlockId = params[1]
        self.controller = controller

        self.target = params[0]['to']
        self.calldata = HexBytes(params[0]['data'])
        self.controller.pending_calls.append(self)
        self._response: Optional[RPCResponse] = None
        demo_logger.info(f'added to queue (cid: {self.cid})')
    
    @property
    def is_complete(self) -> bool:
        return self._response is not None
    
    @property
    def response(self) -> RPCResponse:
        if self._response is None:
            raise ResponseNotReady(self)
        return self._response
    
    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, BatchedCall):
            return False
        return self.cid == __o.cid
    
    def __hash__(self) -> int:
        return self.cid
    
    def __await__(self) -> Generator[Any, None, RPCResponse]:
        return self.wait_for_response().__await__()
    
    async def wait_for_response(self) -> RPCResponse:
        if not self.controller.is_running:
            await self.controller.taskmaster_loop()
        while not self.is_complete:
            await asyncio.sleep(0)
        return self.response

    async def spoof_response(self, data: Optional[Union[bytes, Exception]]) -> RPCResponse:
        """ Sets and returns a spoof rpc response for this BatchedCall instance using data provided by the worker. """
        spoof: RPCResponse
        # If multicall failed, make sync call to get either:
        # - revert details
        # - successful response
        if _call_failed(data):
            data = await self.sync_call()
        
        if isinstance(data, Exception):
            spoof = {"error": _err_response(data)}
        else:
            spoof = {"result": HexBytes(data).hex()}  # type: ignore
        spoof.update({"id": self.cid, "jsonrpc": "dank_mids"})  # type: ignore
        main_logger.debug(f"spoof: {spoof}")
        self._response = spoof
        return spoof

    async def sync_call(self) -> Union[bytes, Exception]:
        """ Used to bypass DankMiddlewareController. """ 
        data = await run_in_subprocess(
            _reattempt_call_and_return_exception, self.target, self.calldata, self.block, self.controller.sync_w3
        )
        # If we were able to get a usable response from single call, add contract to `do_not_batch`.
        if not isinstance(data, Exception):
            self.controller.do_not_batch.add(self.target)
        return data




import asyncio
import threading
from collections import defaultdict
from time import time
from typing import Any, List, Set, Tuple

import multicall
from eth_utils import to_checksum_address
from web3 import Web3
from web3.providers import HTTPProvider
from web3.providers.async_base import AsyncBaseProvider
from web3.types import RPCEndpoint, RPCResponse
from eth_typing import ChecksumAddress


from dank_mids._config import LOOP_INTERVAL
from dank_mids.call import BatchedCall
from dank_mids.loggers import (demo_logger, main_logger, sort_lazy_logger,
                               sort_logger)
from dank_mids.uid import UIDGenerator
from dank_mids.worker import DankWorker

instances: List["DankMiddlewareController"] = []

def _sync_w3_from_async(w3: Web3) -> Web3:
    assert w3.eth.is_async and isinstance(w3.provider, AsyncBaseProvider), "Dank Middleware can only be applied to an asycnhronous Web3 instance."
    sync_provider = HTTPProvider(w3.provider.endpoint_uri)
    sync_w3: Web3 = Web3(provider = sync_provider)
    # Can't pickle middlewares to send to process executor
    sync_w3.middleware_onion.clear()
    sync_w3.provider.middlewares = tuple()
    return sync_w3

def _get_constants_for_network(chainid: int) -> Tuple[ChecksumAddress, Set[ChecksumAddress]]:
    multicall1 = multicall.constants.MULTICALL_ADDRESSES.get(chainid,None)
    multicall2 = multicall.constants.MULTICALL2_ADDRESSES.get(chainid,None)
    return to_checksum_address(multicall2), {to_checksum_address(address) for address in [multicall1,multicall2] if address}


class DankMiddlewareController:
    def __init__(self, w3: Web3) -> None:
        main_logger.info('Dank Middleware initializing... Strap on your rocket boots...')
        self.w3: Web3 = w3
        self.sync_w3 = _sync_w3_from_async(w3)
        self.multicall2, self.do_not_batch = _get_constants_for_network(self.sync_w3.eth.chain_id)
        self.pending_calls: List[BatchedCall] = []
        self.num_pending_calls: int = 0
        self.worker = DankWorker(self)
        self.is_running: bool = False
        self.call_uid = UIDGenerator()
        self._checkpoint: float = time()
        self._instance: int = len(instances)
        instances.append(self)
    
    def __repr__(self) -> str:
        return f"<DankMiddlewareController {self._instance}>"

    async def __call__(self, params: Any) -> RPCResponse:
        call = await self.add_to_queue(params)
        return await call
    
    @property
    def batcher(self):
        return self.worker.batcher
    
    @property
    def pools_closed_lock(self) -> threading.Lock:
        return self.call_uid.lock
    
    async def taskmaster_loop(self) -> None:
        self.is_running = True
        while self.pending_calls:
            await asyncio.sleep(0)
            if (self.loop_is_ready or self.queue_is_full):
                await self.execute_multicall()
        self.is_running = False
    
    async def execute_multicall(self) -> None:
        i = 0
        while self.pools_closed_lock.locked():
            if i // 500 == int(i // 500):
                main_logger.debug('lock is locked')
            i += 1
            await asyncio.sleep(.1)
        with self.pools_closed_lock:
            calls_to_exec = defaultdict(list)
            for call in self.pending_calls:
                calls_to_exec[call.block].append(call)
            self.pending_calls.clear()
            self.num_pending_calls = 0
        demo_logger.info(f'executing multicall (current cid: {self.call_uid.latest})')
        await self.worker.execute_multicall(calls_to_exec)

    @sort_lazy_logger
    def should_batch(self, method: RPCEndpoint, params: Any) -> bool:
        """ Determines whether or not a call should be passed to the DankMiddlewareController. """
        if method != 'eth_call':
            sort_logger.debug(f"bypassed, method is {method}")
            return False
        elif params[0]['to'] in self.do_not_batch:
            sort_logger.debug(f"bypassed, target is in `DO_NOT_BATCH`")
            return False
        return True
    
    async def add_to_queue(self, params: Any) -> "BatchedCall":
        """ Adds a call to the DankMiddlewareContoller's `pending_calls`. """
        return BatchedCall(self, params)

    @property
    def loop_is_ready(self) -> bool:
        return time() - self._checkpoint > LOOP_INTERVAL
    
    @property
    def queue_is_full(self) -> bool:
        return bool(len(self.pending_calls) >= self.batcher.step * 25)



from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict, List, NewType

from web3.types import RPCEndpoint, RPCResponse

if TYPE_CHECKING:
    from dank_mids.call import BatchedCall

BlockId = NewType("BlockId", str)
CallsToExec = Dict[BlockId, List["BatchedCall"]]

# This type alias was introduced in web3 v5.28.0 but we like loose deps here so we recreate instead of import.
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]



import asyncio
import threading
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import aiohttp
from eth_abi import decode_single, encode_single
from eth_utils import function_signature_to_4byte_selector
from hexbytes._utils import to_bytes
from multicall.multicall import batcher
from multicall.utils import gather, run_in_subprocess

from dank_mids.call import BatchedCall
from dank_mids.constants import AIOHTTP_TIMEOUT, OVERRIDE_CODE
from dank_mids.loggers import demo_logger, main_logger
from dank_mids.types import BlockId, CallsToExec
from dank_mids.uid import UIDGenerator

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController

BatchId = Union[int, str]


FOURBYTE = function_signature_to_4byte_selector("tryBlockAndAggregate(bool,(address,bytes)[])")
INPUT_TYPES = "(bool,(address,bytes)[])"
OUTPUT_TYPES = "(uint256,uint256,(bool,bytes)[])"

def bisect(collection: List):
    halfpoint = len(collection) // 2
    left = collection[:halfpoint]
    right = collection[halfpoint:]
    return left, right

class DankWorker:
    """
    Runs a second event loop in a subthread which is used to reduce congestion on the main event loop.
    This allows dank_mids to better communicate with your node while you abuse it with heavy loads.
    """
    def __init__(self, controller: "DankMiddlewareController") -> None:
        self.controller = controller
        self.target = self.controller.multicall2
        self.batcher = batcher
        self.multicall_uid = UIDGenerator()
        self.request_uid = UIDGenerator()
        self.jsonrpc_batch_uid = UIDGenerator()
        self.event_loop = asyncio.new_event_loop()
        self.worker_thread = threading.Thread(target=self.start)
        self.worker_thread.start()
    
    def start(self) -> None:
        """ Runs in worker thread. """
        asyncio.set_event_loop(self.event_loop)
        self.event_loop.run_until_complete(self.loop())
    
    async def loop(self) -> None:
        """ Exits loop when main thread dies, killing worker thread. Runs in worker thread. """
        while threading.main_thread().is_alive():
            await asyncio.sleep(5)
    
    async def execute_multicall(self, calls_to_exec: CallsToExec) -> None:
        """ Runs in main thread. """
        asyncio.run_coroutine_threadsafe(self._execute_multicall(calls_to_exec), self.event_loop).result()
    
    async def _execute_multicall(self, calls_to_exec: CallsToExec) -> None:
        """ Runs in worker thread. """
        full_batches, partial_batches = [], []
        for *batches, partial_batch in [self.batcher.batch_calls(calls, self.batcher.step) for calls in calls_to_exec.values()]:
            full_batches.extend(batches)
            partial_batches.append(partial_batch)
        
        ### Combine multicalls into one or more jsonrpc batches
        partial_batches.sort(key=lambda x: len(x), reverse=True)
        num_calls_in_batch = 0
        jsonrpc_batch = []
        jsonrpc_batches = []
        for partial_batch in partial_batches:
            # This would make the batch too large, lets start a new jsonrpc batch
            if num_calls_in_batch + len(partial_batch) > self.batcher.step:
                if jsonrpc_batch:
                    jsonrpc_batches.append(jsonrpc_batch)
                jsonrpc_batch = [partial_batch]
                num_calls_in_batch = len(partial_batch)
            # This can be added to the current jsonrpc batch
            else:
                jsonrpc_batch.append(partial_batch)
                num_calls_in_batch += len(partial_batch)
                if len(jsonrpc_batch) >= 500:
                    jsonrpc_batches.append(jsonrpc_batch)
                    jsonrpc_batch = []
                    num_calls_in_batch = 0
                
        if jsonrpc_batch:
            jsonrpc_batches.append(jsonrpc_batch)

        coros = [gather(self.process_single_multicall(batch, batch[0].block) for batch in full_batches)]

        if jsonrpc_batches:
            # If the only combined partial batch only contains calls for one block, we can add it to the full batches and process it like any single-block multicall
            if len(jsonrpc_batches) == 1 and len(jsonrpc_batch := jsonrpc_batches[0]) == 1:
                single_batch = jsonrpc_batch[0]
                coros.append(self.process_single_multicall(single_batch, single_batch[0].block))
            else:
                for jsonrpc_batch in jsonrpc_batches:
                    coros.append(self.process_batch_multicall(jsonrpc_batch))
        await gather(coros)
        demo_logger.info('multicall complete')

    def prepare_jsonrpc_batch(self, batches: List[List[BatchedCall]]) -> List[Dict]:
        return [{'jsonrpc': '2.0', 'id': i, 'method': 'eth_call', 'params': self.prepare_multicall_request(batch)} for i, batch in enumerate(batches)]
    
    async def process_batch_multicall(self, batches: List[List[BatchedCall]], jid: Optional[BatchId] = None) -> None:
        """ Runs in worker thread. """
        if jid is None:
            jid = self.jsonrpc_batch_uid.next
        rid = self.request_uid.next
        demo_logger.info(f'executing {sum(len(batch) for batch in batches)} calls for {len(batches)} blocks')
        demo_logger.info(f'request {rid} for jsonrpc batch {jid} starting')
        jsonrpc_batch = self.prepare_jsonrpc_batch(batches)
        try:
            async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
                responses = await session.post(self.controller.w3.provider.endpoint_uri, json=jsonrpc_batch)
                responses = await responses.json()

            # A successful response will be a list
            if isinstance(responses, dict) and 'result' in responses and isinstance(responses['result'], dict) and 'message' in responses['result']:
                raise ValueError(responses['result']['message'])

            await gather([self.process_multicall_response(batch, to_bytes(response)) for batch, response in zip(batches, [response['result'] for response in responses])])
            demo_logger.info(f'request {rid} for jsonrpc batch {jid} complete')
        except Exception as e:
            if "out of gas" in str(e):
                # TODO Remember which contracts/calls are gas guzzlers
                main_logger.debug('out of gas. cut in half, trying again')
            elif any(err in str(e).lower() for err in ["connection reset by peer","request entity too large","server disconnected","execution aborted (timeout = 5s)"]):
                main_logger.debug('dank too loud, trying again')
            else:
                main_logger.warning(f"unexpected exception: {type(e)} {str(e)}")

            chunk0, chunk1 = bisect(batches)
            await gather([
                self.process_batch_multicall(chunk0, str(jid)+"_0") if len(chunk0) > 1 else self.process_single_multicall(chunk0[0], chunk0[0][0].block),
                self.process_batch_multicall(chunk1, str(jid)+"_1") if len(chunk1) > 1 else self.process_single_multicall(chunk1[0], chunk1[0][0].block),
            ])
            demo_logger.info(f'request {rid} for jsonrpc batch {jid} complete')
    
    async def process_single_multicall(self, batch: List[BatchedCall], block: BlockId, bid: Optional[BatchId] = None) -> None:
        """ Runs in worker thread. """
        if bid is None:
            bid = self.multicall_uid.next
        rid = self.request_uid.next
        #demo_logger.info(f'executing {sum(len(batch) for batch in batches)} calls for blocks {[batch[0].block for batch in batches]}')
        demo_logger.info(f'request {rid} for multicall {bid} starting')
        request_args = self.prepare_multicall_request(batch)
        try:
            response = await self.controller.w3.eth.call(*request_args)
            await self.process_multicall_response(batch, response)
            demo_logger.info(f'request {rid} for multicall {bid} complete')
        except Exception as e:
            if len(batch) == 1:
                await batch[0].spoof_response(e)
                return
            elif "out of gas" in str(e):
                # TODO Remember which contracts/calls are gas guzzlers
                main_logger.debug('out of gas. cut in half, trying again')
            elif any(err in str(e).lower() for err in ["connection reset by peer","request entity too large","server disconnected","execution aborted (timeout = 5s)"]):
                main_logger.debug('dank too loud, trying again')
                # We need this check because one of the other multicalls in a batch might have already reduced `self.batcher.step`
                if (new_step := round(num_calls * 0.99) if (num_calls := len(batch)) >= 100 else num_calls - 1) < self.batcher.step:
                    old_step = self.batcher.step
                    self.batcher.step = new_step
                    main_logger.warning(f'Multicall batch size reduced from {old_step} to {new_step}. The failed batch had {num_calls} calls.')
            else:
                main_logger.warning(f"unexpected exception: {type(e)} {str(e)}")

            chunk0, chunk1 = bisect(batch)
            await gather([
                self.process_single_multicall(chunk0, block, str(bid)+"_0"),
                self.process_single_multicall(chunk1, block, str(bid)+"_1"),
            ])
            demo_logger.info(f'request {rid} for multicall {bid} complete')
    
    def prepare_multicall_request(self, batch: List[BatchedCall]) -> List[Union[Dict, str]]:
        fn_args = [False, [[call.target, call.calldata] for call in batch]]
        calldata = FOURBYTE + encode_single(INPUT_TYPES, fn_args)
        return [{'to': self.target, 'data': '0x' + calldata.hex()}, batch[0].block, {self.target: {'code': OVERRIDE_CODE}}]

    async def process_multicall_response(self, batch: List[BatchedCall], response: bytes) -> None:
        _, _, response = await run_in_subprocess(decode_single, OUTPUT_TYPES, response)
        await gather([call.spoof_response(data) for call, (_, data) in zip(batch, response)])
